
// ****************************************************************************
// File: "Code Analyzer.cpp"
// Desc: Binary code hook analyze support
// http://www.sirmabus.macromonkey.com/?p=124
//
// ****************************************************************************
#include "stdafx.h"
#include "Code Analyzer.h"

#define IsAlignmentByte(_Byte) ((_Byte == INSTR_INT3) || (_Byte == INSTR_NOP))


// Get instruction relocation information for a given area of code space
// Input: 
//	pCodeBytes  = Target code space
//  iCodeLength = Size to relocate.
//  rInfo       = Instruction info container.
//
// Returns TRUE on success
BOOL CODE::GetRelocInfo(const PBYTE pCodeBytes, int iCodeLength, RELOCINFO &rInfo)
{
	// Free up allocation and reset input struct so it can be reused
	rInfo.Free();
	
	__try
	{	
		PBYTE pCode = (PBYTE) pCodeBytes;
		int   iLen  = iCodeLength;	
		BOOL  bHitReturn = FALSE;

		while((iLen > 0) && !bHitReturn)
		{
			// Disassemble next instruction
			if((rInfo.pDissasm = (LPHDE32S) _aligned_realloc(rInfo.pDissasm, (sizeof(HDE32S) * (rInfo.iInstCount + 1)), 16)) == NULL)
				return(FALSE);
			LPHDE32S pInfo = &rInfo.pDissasm[rInfo.iInstCount++];
			UINT uInstLen  = hde32_disasm(pCode, pInfo);
			rInfo.iRoundedLen += (int) uInstLen;
			pCode += uInstLen; iLen -= (int) uInstLen;
						
			// On disassemble error, flag and skip instruction
			if(pInfo->flags & F_ERROR)
			{
				rInfo.Flags |= HAS_INST_ERROR;				
			}
			else
			{	
				// Two byte opcodes of interest
				if(pInfo->opcode == INSTR_TWOBYTE_0F)
				{
					// rel16/32 jumps
					if((pInfo->opcode2 & 0xF0) == INSTR_REL16_32JCC_START)
						rInfo.Flags |= HAS_REL_INST;
				}
				else
				{					
					// Hit a jmp or call with a relative offset?
					BYTE OpCode = pInfo->opcode;
					if(pInfo->flags & F_RELATIVE)
					{				
						// rel8 jumps
						if((OpCode == INSTR_REL8JMP) || ((OpCode & 0xF0) == INSTR_REL8JCC_START) || ((OpCode >= INSTR_LOOPNZ) && (OpCode <= INSTR_REL8JCX)))						
							rInfo.Flags |= HAS_REL_INST;						
						else
						// rel16/32 jump
						if(OpCode == INSTR_REL16_32JMP)					
							rInfo.Flags |= HAS_REL_INST;					
						else
						// rel16/32 call
						if(OpCode == INSTR_REL16_32CALL)					
							rInfo.Flags |= HAS_REL_INST;					
					}
					else
					// Look for possible end of function
					{				
						// Hit a return opcode, or an int3?
						// Int3's often used as alignment bytes tell us this is the end of the function.
						// Perhaps the function is a single "jmp".						
						switch(pInfo->opcode)
						{
							// Once we hit a return instruction, we assume it's the end of the function
							// This can't always be right of course
							case INSTR_RETN: case INSTR_RET: case INSTR_RETFN: case INSTR_RETF:
							{
								if(bHitReturn)
									break;
								else
									bHitReturn = TRUE;

								/*
								// Disassemble next instruction
								if((rInfo.pDissasm = (LPHDE32S) _aligned_realloc(rInfo.pDissasm, (sizeof(HDE32S) * (rInfo.iInstCount + 1)), 16)) == NULL)
									return(FALSE);
								LPHDE32S pInfo = &rInfo.pDissasm[rInfo.iInstCount++];
								UINT uInstLen  = hde32_disasm(pCode, pInfo);
								rInfo.iRoundedLen += (int) uInstLen;
								pCode += uInstLen; iLen -= (int) uInstLen;
								*/
							}

							// Pretty much the same with alignment instructions
							// We assume, we are in alignment bytes after a return, or other ending branch
							case INSTR_INT3: case INSTR_NOP:
							{
								/*
									This could be the end of the function, just the end of a block inside a function.									
									We'll just assume for now that that it's the end and look for following alignment
									bytes if we need to.							
								*/

								// Walk though the following instructions for possible trailing alignment bytes that 
								// would most likely still be ok for us to overwrite.						
								while(iLen > 0)
								{
									HDE32S tInfo;
									UINT uInstLen = hde32_disasm(pCode, &tInfo);							
									
									if((uInstLen != 1) || (tInfo.flags & F_ERROR) || !IsAlignmentByte(tInfo.opcode))
									{
										rInfo.Flags |= NOT_ENOUGH_ROOM;
										break;
									}
									else
									{
										// Store struct, flag, and continue
										if((rInfo.pDissasm = (LPHDE32S) _aligned_realloc(rInfo.pDissasm, (sizeof(HDE32S) * (rInfo.iInstCount + 1)), 16)) == NULL)
											return(FALSE);
										rInfo.pDissasm[rInfo.iInstCount++] = tInfo;
										rInfo.Flags |= WILL_OVERFLOW;
										rInfo.iRoundedLen += (int) uInstLen;								
										pCode += uInstLen; iLen -= (int) uInstLen;
									}							
								};															
							}
							break;

							// If we found anything else after a return, then assume we ran out of room
							default:
							if(bHitReturn)
							{
								rInfo.Flags |= NOT_ENOUGH_ROOM;
							}
							break;
						};
					}
				}
			}
		}; // while((iLen > 0) && !bHitReturn)
	}
	// Catch code memory access errors
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		return(FALSE);
	}

	return(TRUE);
}