
// ****************************************************************************
// File: Main.cpp 
// Desc: DLL export code scanner test
//       http://www.sirmabus.macromonkey.com/?p=124
//
// ****************************************************************************
#include "stdafx.h"
#include "Code Analyzer.h"

#define HOOK_TEST_SIZE 5 // Hook byte test size
#define DUMP_SIZE 8		 // Instruction dump length


// Stats
UINT g_uDLLs         = 0;	// Valid DLLs examined
UINT g_uExports		 = 0;	// Total exports examined
UINT g_uExpRedundant = 0;	// Count of redundant export addresses
UINT g_uInstError	 = 0;	// Instruction decode error instances
UINT g_uRelInst      = 0;	// Relative offset instruction instances
UINT g_uRel32Stub    = 0;	// Has relative 32bit branch on function entry
UINT g_uOverflow     = 0;	// Overflowing into alignment byte stub instances
UINT g_uNoRoom       = 0;	// No room for HOOK_TEST_SIZE cases


// PE header stuff
#define MakePtr(_cast, _ptr, _addValue) reinterpret_cast<_cast>((UINT) (_ptr) + (UINT) (_addValue))
#define GetImgDirEntryRVA(pNTHdr,  IDE) (pNTHdr->OptionalHeader.DataDirectory[IDE].VirtualAddress)
#define GetImgDirEntrySize(pNTHdr, IDE) (pNTHdr->OptionalHeader.DataDirectory[IDE].Size)

template <class T> PIMAGE_SECTION_HEADER GetEnclosingSectionHeader(UINT rva, T *pNTHeader)
{
	PIMAGE_SECTION_HEADER ptSecHdr = IMAGE_FIRST_SECTION(pNTHeader);
	for(UINT i = 0; i < pNTHeader->FileHeader.NumberOfSections; i++, ptSecHdr++)
	{		
		UINT uSize = ptSecHdr->Misc.VirtualSize;
		if(!uSize)
			uSize = ptSecHdr->SizeOfRawData;

		// Is the RVA within this section?
		if((rva >= ptSecHdr->VirtualAddress) && (rva < (ptSecHdr->VirtualAddress + uSize)))
			return(ptSecHdr);
	}

	return(NULL);
}

// Get mapped offset to a mapped section
template <class T, class U> int GetMappedOffset(T *pNTHeader, U *ptSectionHdr, PVOID pImageBase)
{
	PIMAGE_SECTION_HEADER pSectionHdr = IMAGE_FIRST_SECTION(pNTHeader);
	int iOffset = ((int) pSectionHdr - (int) pImageBase);

	for(UINT i = 0; i < pNTHeader->FileHeader.NumberOfSections; i++, pSectionHdr++)
	{
		if(pSectionHdr->Name == ptSectionHdr->Name)
		{
			iOffset = ((int) pSectionHdr->VirtualAddress - (int) pSectionHdr->PointerToRawData);
			break;
		}
	}

	return(iOffset);
}

template <class T> LPVOID GetPtrFromRVA(UINT rva, T *pNTHeader, PVOID pImageBase)
{
	if(PIMAGE_SECTION_HEADER pSectionHdr = GetEnclosingSectionHeader(rva, pNTHeader))
	{				
		int iDelta = (int) (pSectionHdr->VirtualAddress - pSectionHdr->PointerToRawData);
		return((PVOID) ((UINT) pImageBase + rva - iDelta));
	}

	return(NULL);
}

// Returns TRUE if PE file has a known compressed section
BOOL HasCompressedSection(PIMAGE_NT_HEADERS ptNTHeader)
{
	// TODO: Add more section names..
	const static LPCSTR aSectionName[] =
	{
		"UPX0",".adata"
	};

	//printf("=========== Sections ==========\n");
	PIMAGE_SECTION_HEADER ptSecHdr = IMAGE_FIRST_SECTION(ptNTHeader);
	for(UINT i = 0; i < ptNTHeader->FileHeader.NumberOfSections; i++, ptSecHdr++)
	{	
		//printf("[%u] \"%s\"\n", i, ptSecHdr->Name);
		if(ptSecHdr->Name[0])
		{
			for(UINT j = 0; j < (sizeof(aSectionName) / sizeof(LPCSTR)); j++)
			{
				if(strncmp(aSectionName[j], (LPCSTR) ptSecHdr->Name, sizeof(ptSecHdr->Name)) == 0)
					return(TRUE);
			}
		}
	}
	//printf("\n");

	return(FALSE);
}

// Returns TRUE if address is in any known data section
BOOL IsDataSection(PIMAGE_NT_HEADERS ptNTHeader, UINT uAddress)
{
	const static LPCSTR aSectionName[] =
	{
		"data","bss",
	};	

	if(PIMAGE_SECTION_HEADER ptSecHdr = GetEnclosingSectionHeader(uAddress, ptNTHeader))
	{
		if(ptSecHdr->Name[0])
		{			
			for(UINT i = 0; i < (sizeof(aSectionName) / sizeof(LPCSTR)); i++)
			{
				if(strstr((LPCSTR) ptSecHdr->Name, aSectionName[i]))
					return(TRUE);
			}
		}
	}

	return(FALSE);
}



// Read in a whole DLL
// You free up the buffer passed though lpMemory
long RawReadDLL(LPCTSTR fn, PBYTE *lpMemory)
{	
	long fSize = 0;
	if(FILE *fp = fopen(fn, "rb"))
	{
		fseek(fp, 0, SEEK_END);
		fSize = ftell(fp);
		fseek(fp,0,SEEK_SET);

		if(fSize > 0)
		{
			*lpMemory = (PBYTE) _aligned_malloc(fSize, 32);
			if(*lpMemory)
			{
				if(fread(*lpMemory, fSize, 1, fp) != 1)
					fSize = 0;
			}
			else
				fSize = 0;
		}

		fclose(fp);
	}	
	return(fSize);
}


// Log formated text to file and console
void Log(FILE *hLogFile, LPCTSTR pszFormat, ...)
{
	if(pszFormat)
	{	
		va_list vl;
		char szBuffer[4096];
		va_start(vl, pszFormat);
		_vsntprintf(szBuffer, (sizeof(szBuffer) - 1), pszFormat, vl);
		szBuffer[sizeof(szBuffer) - 1] = 0;
		va_end(vl);

		#ifdef _DEBUG
		puts(szBuffer);
		#endif
		fputs(szBuffer, hLogFile);
		fputs("\n", hLogFile);
	}
}


// Verify we can read a structure at a given address
#define IsGoodStruct(_ptr, _struct) !IsBadReadPtr((const VOID*) _ptr, sizeof(_struct))
#define IsGoodString(_ptr, _maxlen) !IsBadStringPtrA(_ptr, _maxlen) 

// Process DLL file
ALIGN(16) CODE::RELOCINFO CodeInfo;
void ProcessFile(LPCTSTR pszFile, FILE *hLogFile)
{	
	Log(hLogFile, "\n\nDLL: \"%s\"", pszFile);
			
	// Read in DLL	
	PBYTE pBase = NULL;
	if(RawReadDLL(pszFile, &pBase) > sizeof(PIMAGE_DOS_HEADER))
	{
		/*
			Can get hairy here; And why a lot of generic PE parsers fail
			The DLL might be packed, etc.
			Need to check each step..
		*/

		// Should be a DLL PE header
		PIMAGE_NT_HEADERS        ptNTHeader  = NULL;
		PIMAGE_EXPORT_DIRECTORY  ptExportDir = NULL;
		PIMAGE_OPTIONAL_HEADER32 ptOptHeader = NULL;

		PIMAGE_DOS_HEADER ptDosHeader = (PIMAGE_DOS_HEADER) pBase;		
		if(IsGoodStruct(ptDosHeader, IMAGE_DOS_HEADER) && (ptDosHeader->e_magic == IMAGE_DOS_SIGNATURE))
		{	
			ptNTHeader  = MakePtr(PIMAGE_NT_HEADERS, pBase, ptDosHeader->e_lfanew);
			if(IsGoodStruct(ptNTHeader, IMAGE_NT_HEADERS) && (ptNTHeader->Signature == LOWORD(IMAGE_NT_SIGNATURE)))
			{
				PIMAGE_FILE_HEADER ptFileHeader = &ptNTHeader->FileHeader;
				if(IsGoodStruct(ptFileHeader, IMAGE_FILE_HEADER))
				{
					if(ptFileHeader->Characteristics & IMAGE_FILE_DLL)
					{
						ptOptHeader = &ptNTHeader->OptionalHeader;
						if(IsGoodStruct(ptOptHeader, IMAGE_OPTIONAL_HEADER))
						{							
							if(ptOptHeader->Magic == IMAGE_NT_OPTIONAL_HDR_MAGIC)
							{				
								// Only interested with DLLs that have an export section							
								if(UINT uExportsStartRVA = GetImgDirEntryRVA(ptNTHeader, IMAGE_DIRECTORY_ENTRY_EXPORT))
								{
									if(!HasCompressedSection(ptNTHeader))
									{
										ptExportDir = (PIMAGE_EXPORT_DIRECTORY) GetPtrFromRVA(uExportsStartRVA, ptNTHeader, pBase);
									}
									//else
									//	printf("\n* \"%s\" is compressed/encrypted.\n", pszFile);							
								}
							}
						}

					}
				}
			}
		}

		if(ptExportDir)
		{
			// Stored name
			// If it's not present the file is packed or otherwise invalid
			LPCTSTR pszStoredName = (LPCTSTR) GetPtrFromRVA(ptExportDir->Name, ptNTHeader, pBase);
			if(IsGoodString(pszStoredName, _MAX_FNAME))
			{
				Log(hLogFile, "Name: \"%s\"", pszStoredName);										
			
				// Walk through exports
				if(UINT uNameCount = ptExportDir->NumberOfNames)
				{
					UINT  uFuncCount = ptExportDir->NumberOfFunctions;
					PUINT pFunctions = (PUINT) GetPtrFromRVA(ptExportDir->AddressOfFunctions,    ptNTHeader, pBase);
					PWORD pOrdinals  = (PWORD) GetPtrFromRVA(ptExportDir->AddressOfNameOrdinals, ptNTHeader, pBase);
					PUINT pNames     = (PUINT) GetPtrFromRVA(ptExportDir->AddressOfNames,        ptNTHeader, pBase);										
					_ASSERT(uFuncCount && pFunctions && pOrdinals && pNames);
					Log(hLogFile, "Names: %d, functions: %d", uNameCount, uFuncCount);
					g_uDLLs++;

					// A local array to track redundant addresses
					// Some exports will point to the same addresss
					// We don't want to examine the same export more then once
					UINT  uVisited = 0;
					PUINT pVisited = (PUINT) _aligned_malloc((uNameCount * sizeof(UINT)), 16);
					ZeroMemory(pVisited, (uNameCount * sizeof(UINT)));
				
					UINT uBaseOrdinal = ptExportDir->Base;										
					for(UINT i = 0; i < uNameCount; i++)
					{
						if(UINT uEntryPointRVA = pFunctions[pOrdinals[i]])
						{
							// Skip if it's all ready been examined
							BOOL bExists = FALSE;
							{					
								PUINT pPtr  = pVisited;
								UINT uCount = uVisited;
								while(uCount--)
								{
									if(*pPtr++ == uEntryPointRVA)
									{
										g_uExpRedundant++;
										bExists = TRUE;
										break;
									}
								};

								// Add it if it doesn't exist
								if(!bExists)								
									pVisited[uVisited++] = uEntryPointRVA;								
							}

							if(!bExists)
							{		
								if(!IsDataSection(ptNTHeader, uEntryPointRVA))
								{								
									if(UINT pRVA = (UINT) GetPtrFromRVA(uEntryPointRVA, ptNTHeader, pBase))
									{													
										LPCTSTR pszName  = (LPCTSTR) GetPtrFromRVA(pNames[i], ptNTHeader, pBase);
										UINT    uVirtual = (ptOptHeader->ImageBase + uEntryPointRVA);
										Log(hLogFile, "\n=======================================================================");								
										Log(hLogFile, "[%u] \"%s\" %d %08X", i, pszName, (pOrdinals[i] + uBaseOrdinal), uVirtual);									
										
										/*
										if(uVirtual == 0x7C880311)
										{
											_asm int 3;									
										}
										*/

										// Disassemble the first few instructions and log them												
										// ----------------------------------------------------------------
										__try
										{
											PBYTE pInput = (PBYTE) pRVA;
											UINT uOffset = 0;

											for(UINT j = 0; j < DUMP_SIZE; j++)
											{
												// Get instruction length first since DStorm might fail on invalid sized instructions															
												HDE32S tInfo;
												UINT uInstLen = hde32_disasm(pInput, &tInfo);
												if(!uInstLen) break;

												_DecodedInst tInstInfo;
												UINT uCount = 0;
												if(distorm_decode32(uOffset, pInput, uInstLen, Decode32Bits, &tInstInfo, 1, &uCount) == DECRES_SUCCESS)																													
													Log(hLogFile, " [%02X] %s\t%s %s", tInstInfo.offset, tInstInfo.instructionHex.p, tInstInfo.mnemonic.p, tInstInfo.operands.p);															
												else
												{
													Log(hLogFile, "* distorm_decode32() fail!");
													break;
												}

												pInput  += uInstLen;
												uOffset += uInstLen;
											}
										}
										__except(1)
										{
											Log(hLogFile, "* distorm_decode32() exception!");														
										}


										// Analyze code with simulated hook stub placing
										// ----------------------------------------------------------------
										__try
										{									
											if(CODE::GetRelocInfo((const PBYTE) pRVA, HOOK_TEST_SIZE, CodeInfo))
											{							
												if(CodeInfo.Flags & CODE::HAS_INST_ERROR)
												{		
													Log(hLogFile, "*** CODE::HAS_INST_ERROR\n");
													g_uInstError++;
												}

												if(CodeInfo.Flags & CODE::HAS_REL_INST)
												{
													// Only on entry?
													// We treat a 32bit relative first jmp or call uniquely
													// With a JMP5 hook we can just replace it
													BOOL bIsRel32Entry = FALSE;
													#if (HOOK_TEST_SIZE == 5)
													if(!(CodeInfo.Flags & CODE::NOT_ENOUGH_ROOM))
													{
														// First is relative and 32bit offset?
														if((CodeInfo.pDissasm[0].flags & (F_RELATIVE | F_IMM32)) == (F_RELATIVE | F_IMM32))
														{
															int k = 1;
															for(; k < CodeInfo.iInstCount; k++)
															{
																// Stop if we find another relative instruction
																if(CodeInfo.pDissasm[0].flags & F_RELATIVE)
																	break;																		
															}	
															bIsRel32Entry = (k == CodeInfo.iInstCount);
														}
													}																															
													#endif

													if(bIsRel32Entry)
													{
														Log(hLogFile, "** Relative 32bit entry branch.\n");
														g_uRel32Stub++;
													}
													else
													{
														Log(hLogFile, "*** CODE::HAS_REL_INST\n");
														g_uRelInst++;
													}
												}

												if(CodeInfo.Flags & CODE::WILL_OVERFLOW)
												{
													Log(hLogFile, "** CODE::WILL_OVERFLOW\n");
													g_uOverflow++;
												}

												if(CodeInfo.Flags & CODE::NOT_ENOUGH_ROOM)
												{
													Log(hLogFile, "*** CODE::NOT_ENOUGH_ROOM\n");
													g_uNoRoom++;
												}
											}
										}
										__except(1)
										{
											Log(hLogFile, "** CODE::GetRelocInfo() exception!");
										}

										g_uExports++;
									}
									else
									{
										_ASSERT(FALSE);
									}
								}
							}
						}
						else
						{
							_ASSERT(FALSE);
						}
					}

					// Free visited address array
					_aligned_free(pVisited);					
				}
			}
		}
		
		// Free DLL memory
		_aligned_free(pBase);
	}
	else	
		printf("** Read error, and, or, not a valid DLL file! **\n");	
}


// ===================================================
int main(int argc, char *argv[])
{
	// By default use all DLL in current directory
	char szSearchPath[MAX_PATH] = {"*.DLL"};
	if(argc == 2)
		strcpy(szSearchPath, argv[1]);
	
	printf("Working..");

	FILE *hLogFile = fopen("log.txt", "wb");
	_ASSERT(hLogFile);

	WIN32_FIND_DATA tFileInfo = {0};
	HANDLE hFile = FindFirstFile(szSearchPath, &tFileInfo);
	if(hFile != INVALID_HANDLE_VALUE) 
	{
		ProcessFile(tFileInfo.cFileName, hLogFile);			

		while(FindNextFile(hFile, &tFileInfo))
		{
			ProcessFile(tFileInfo.cFileName, hLogFile);			
		};

		FindClose(hFile);
	}
	
	Log(hLogFile, "\n\n========== Stats =========");
	Log(hLogFile, "             DLLs: %u", g_uDLLs);
	Log(hLogFile, "    Total exports: %u", g_uExports);
	Log(hLogFile, "Redundant exports: %u\t\t%.1f%%", g_uExpRedundant, (((double) g_uExpRedundant / (double) g_uExports) * 100.0));		
	Log(hLogFile, "    Decode errors: %u\t\t%.1f%%", g_uInstError, (((double) g_uInstError / (double) g_uExports) * 100.0));
	Log(hLogFile, "Relative branches: %u\t\t%.1f%%", g_uRelInst, (((double) g_uRelInst / (double) g_uExports) * 100.0));
	Log(hLogFile, "   Branch entries: %u\t\t%.1f%%", g_uRel32Stub, (((double) g_uRel32Stub / (double) g_uExports) * 100.0));		
	Log(hLogFile, "     Over follows: %u\t\t%.1f%%", g_uOverflow, (((double) g_uOverflow / (double) g_uExports) * 100.0));
	Log(hLogFile, "      Out of room: %u\t\t%.1f%%", g_uNoRoom, (((double) g_uNoRoom / (double) g_uExports) * 100.0)); 

	fclose(hLogFile);

	return(0);
}

